@@ -2,10 +2,10 @@ import importlib
 import inspect
 import logging
 import pkgutil
-import warnings
-from typing import Text, List, Dict, Any, Type, Union, Callable, Optional, Set
-from collections import namedtuple
 import typing
+import warnings
+from typing import Text, List, Dict, Any, Type, Union, Callable, Optional, Set, cast
+from collections import namedtuple
 import types
 import sys
 import os
@@ -14,6 +14,9 @@ from rasa_sdk.interfaces import Tracker, ActionNotFoundException, Action
 
 from rasa_sdk import utils
 
+if typing.TYPE_CHECKING:  # pragma: no cover
+    from rasa_sdk.types import ActionCall
+
 logger = logging.getLogger(__name__)
 
 
@@ -22,7 +25,7 @@ class CollectingDispatcher:
 
     def __init__(self) -> None:
 
-        self.messages = []
+        self.messages: List[Dict[Text, Any]] = []
 
     def utter_message(
         self,
@@ -147,12 +150,13 @@ TimestampModule = namedtuple("TimestampModule", ["timestamp", "module"])
 
 class ActionExecutor:
     def __init__(self) -> None:
-        self.actions = {}
+        self.actions: Dict[Text, Callable] = {}
         self._modules: Dict[Text, TimestampModule] = {}
-        self._loaded: Set[Type] = set()
+        self._loaded: Set[Type[Action]] = set()
 
-    def register_action(self, action: Union[Type["Action"], "Action"]) -> None:
+    def register_action(self, action: Union[Type[Action], Action]) -> None:
         if inspect.isclass(action):
+            action = cast(Type[Action], action)
             if action.__module__.startswith("rasa."):
                 logger.warning(f"Skipping built in Action {action}.")
                 return
@@ -208,7 +212,7 @@ class ActionExecutor:
         if not getattr(package, "__path__", None):
             return
 
-        for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
+        for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):  # type: ignore  # mypy issue #1422
             full_name = package.__name__ + "." + name
             self._import_module(full_name)
 
@@ -255,8 +259,7 @@ class ActionExecutor:
         self._register_all_actions()
 
     def _register_all_actions(self) -> None:
-        """Scan for all user subclasses of `Action`, and register them.
-        """
+        """Scan for all user subclasses of `Action`, and register them."""
         import inspect
 
         actions = utils.all_subclasses(Action)
@@ -337,19 +340,20 @@ class ActionExecutor:
     @staticmethod
     def validate_events(events: List[Dict[Text, Any]], action_name: Text):
         validated = []
-        for e in events:
-            if isinstance(e, dict):
-                if not e.get("event"):
+        for event in events:
+            if isinstance(event, dict):
+                if not event.get("event"):
                     logger.error(
                         f"Your action '{action_name}' returned an action dict "
                         "without the `event` property. Please use "
                         "the helpers in `rasa_sdk.events`! Event will"
-                        f"be ignored! Event: {e}"
+                        f"be ignored!"
                     )
+                    logger.debug(f"Contents of ignored event: '{event}'")
                 else:
-                    validated.append(e)
-            elif type(e).__module__ == "rasa.core.events":
-                logger.warning(
+                    validated.append(event)
+            elif type(event).__module__ == "rasa.core.events":
+                warnings.warn(
                     "Your action should not return Rasa actions within the "
                     "SDK. Instead of using events from "
                     "`rasa.core.events`, you should use the ones "
@@ -357,16 +361,17 @@ class ActionExecutor:
                     "We will try to make this work, but this "
                     "might go wrong!"
                 )
-                validated.append(e.as_dict())
+                validated.append(event.as_dict())  # type: ignore
             else:
                 logger.error(
                     f"Your action's '{action_name}' run method returned an invalid "
-                    f"event. Event will be ignored. Event: '{e}'."
+                    f"event. Event will be ignored."
                 )
+                logger.debug(f"Contents of ignored event: '{event}'")
                 # we won't append this to validated events -> will be ignored
         return validated
 
-    async def run(self, action_call: Dict[Text, Any]) -> Optional[Dict[Text, Any]]:
+    async def run(self, action_call: "ActionCall") -> Optional[Dict[Text, Any]]:
         from rasa_sdk.interfaces import Tracker
 
         action_name = action_call.get("next_action")
@@ -376,7 +381,7 @@ class ActionExecutor:
             if not action:
                 raise ActionNotFoundException(action_name)
 
-            tracker_json = action_call.get("tracker")
+            tracker_json = action_call["tracker"]
             domain = action_call.get("domain", {})
             tracker = Tracker.from_dict(tracker_json)
             dispatcher = CollectingDispatcher()
@@ -393,5 +398,6 @@ class ActionExecutor:
             validated_events = self.validate_events(events, action_name)
             logger.debug(f"Finished running '{action_name}'")
             return self._create_api_response(validated_events, dispatcher.messages)
-        else:
-            logger.warning("Received an action call without an action.")
+
+        logger.warning("Received an action call without an action.")
+        return None
