@@ -1,8 +1,13 @@
 import copy
 import logging
 import typing
+import warnings
 from typing import Any, Dict, Iterator, List, Optional, Text
 
+if typing.TYPE_CHECKING:  # pragma: no cover
+    from rasa_sdk.types import DomainDict, TrackerState
+
+
 logger = logging.getLogger(__name__)
 
 ACTION_LISTEN_NAME = "action_listen"
@@ -12,17 +17,17 @@ class Tracker:
     """Maintains the state of a conversation."""
 
     @classmethod
-    def from_dict(cls, state: Dict[Text, Any]) -> "Tracker":
+    def from_dict(cls, state: "TrackerState") -> "Tracker":
         """Create a tracker from dump."""
 
         return Tracker(
-            state.get("sender_id"),
+            state["sender_id"],
             state.get("slots", {}),
             state.get("latest_message", {}),
-            state.get("events"),
-            state.get("paused"),
+            state.get("events", []),
+            state.get("paused", False),
             state.get("followup_action"),
-            state.get("active_form", {}),
+            state.get("active_loop", state.get("active_form", {})),
             state.get("latest_action_name"),
         )
 
@@ -34,7 +39,7 @@ class Tracker:
         events: List[Dict[Text, Any]],
         paused: bool,
         followup_action: Optional[Text],
-        active_form: Optional[Text],
+        active_loop: Dict[Text, Any],
         latest_action_name: Optional[Text],
     ) -> None:
         """Initialize the tracker."""
@@ -55,9 +60,17 @@ class Tracker:
         #                   "entities": UserUttered.entities,
         #                   "text": text}
         self.latest_message = latest_message if latest_message else {}
-        self.active_form = active_form
+        self.active_loop = active_loop
         self.latest_action_name = latest_action_name
 
+    @property
+    def active_form(self) -> Dict[Text, Any]:
+        warnings.warn(
+            "Use of `active_form` is deprecated. Please use `active_loop insteaad.",
+            DeprecationWarning,
+        )
+        return self.active_loop
+
     def current_state(self) -> Dict[Text, Any]:
         """Return the current tracker state as an object."""
 
@@ -74,7 +87,7 @@ class Tracker:
             "paused": self.is_paused(),
             "events": self.events,
             "latest_input_channel": self.get_latest_input_channel(),
-            "active_form": self.active_form,
+            "active_loop": self.active_loop,
             "latest_action_name": self.latest_action_name,
         }
 
@@ -88,18 +101,39 @@ class Tracker:
         if key in self.slots:
             return self.slots[key]
         else:
-            logger.info(f"Tried to access non existent slot '{key}'")
+            logger.info(f"Tried to access non existent slot '{key}'.")
             return None
 
-    def get_latest_entity_values(self, entity_type: Text) -> Iterator[Text]:
-        """Get entity values found for the passed entity name in latest msg.
+    def get_latest_entity_values(
+        self,
+        entity_type: Text,
+        entity_role: Optional[Text] = None,
+        entity_group: Optional[Text] = None,
+    ) -> Iterator[Text]:
+        """Get entity values found for the passed entity type and optional role and
+        group in latest message.
 
         If you are only interested in the first entity of a given type use
         `next(tracker.get_latest_entity_values("my_entity_name"), None)`.
-        If no entity is found `None` is the default result."""
+        If no entity is found `None` is the default result.
+
+        Args:
+            entity_type: the entity type of interest
+            entity_role: optional entity role of interest
+            entity_group: optional entity group of interest
+
+        Returns:
+            List of entity values.
+        """
 
         entities = self.latest_message.get("entities", [])
-        return (x.get("value") for x in entities if x.get("entity") == entity_type)
+        return (
+            x.get("value")
+            for x in entities
+            if x.get("entity") == entity_type
+            and (entity_group is None or x.get("group") == entity_group)
+            and (entity_role is None or x.get("role") == entity_role)
+        )
 
     def get_latest_input_channel(self) -> Optional[Text]:
         """Get the name of the input_channel of the latest UserUttered event"""
@@ -145,7 +179,7 @@ class Tracker:
             copy.deepcopy(self.events),
             self._paused,
             self.followup_action,
-            self.active_form,
+            self.active_loop,
             self.latest_action_name,
         )
 
@@ -175,20 +209,21 @@ class Tracker:
 
         def undo_till_previous(event_type: Text, done_events: List[Dict[Text, Any]]):
             """Removes events from `done_events` until the first
-                occurrence `event_type` is found which is also removed."""
+            occurrence `event_type` is found which is also removed."""
             # list gets modified - hence we need to copy events!
             for e in reversed(done_events[:]):
                 del done_events[-1]
                 if e["event"] == event_type:
                     break
 
-        applied_events = []
+        applied_events: List[Dict[Text, Any]] = []
         for event in self.events:
-            if event.get("name") == "restart":
+            event_type = event.get("event")
+            if event_type == "restart":
                 applied_events = []
-            elif event.get("name") == "undo":
+            elif event_type == "undo":
                 undo_till_previous("action", applied_events)
-            elif event.get("name") == "rewind":
+            elif event_type == "rewind":
                 # Seeing a user uttered event automatically implies there was
                 # a listen event right before it, so we'll first rewind the
                 # user utterance, then get the action right before it (also removes
@@ -199,6 +234,29 @@ class Tracker:
                 applied_events.append(event)
         return applied_events
 
+    def slots_to_validate(self) -> Dict[Text, Any]:
+        """Get slots which were recently set.
+
+        This can e.g. be used to validate form slots after they were extracted.
+
+        Returns:
+            A mapping of extracted slot candidates and their values.
+        """
+
+        slots: Dict[Text, Any] = {}
+
+        for event in reversed(self.events):
+            # The `FormAction` in Rasa Open Source will append all slot candidates
+            # at the end of the tracker events.
+            if event["event"] == "slot":
+                slots[event["name"]] = event["value"]
+            else:
+                # Stop as soon as there is another event type as this means that we
+                # checked all potential slot candidates.
+                break
+
+        return slots
+
 
 class Action:
     """Next action to be taken in response to a dialogue state."""
@@ -209,16 +267,17 @@ class Action:
         raise NotImplementedError("An action must implement a name")
 
     async def run(
-        self, dispatcher, tracker: Tracker, domain: Dict[Text, Any]
+        self,
+        dispatcher,
+        tracker: Tracker,
+        domain: "DomainDict",
     ) -> List[Dict[Text, Any]]:
         """Execute the side effects of this action.
 
         Args:
             dispatcher: the dispatcher which is used to
                 send messages back to the user. Use
-                ``dipatcher.utter_message()`` or any other
-                ``rasa_sdk.executor.CollectingDispatcher``
-                method.
+                ``dispatcher.utter_message()`` for sending messages.
             tracker: the state tracker for the current
                 user. You can access slot values using
                 ``tracker.get_slot(slot_name)``, the most recent user message
@@ -238,7 +297,7 @@ class Action:
 
 class ActionExecutionRejection(Exception):
     """Raising this exception will allow other policies
-        to predict another action"""
+    to predict another action"""
 
     def __init__(self, action_name: Text, message: Optional[Text] = None) -> None:
         self.action_name = action_name
